from math import floor, sqrt
import random
import matplotlib.pyplot as plt
import sibc.montgomery as mont
from sibc.constants import parameters
from montgomery_new import MontgomeryIsogenyNew

class IsogenyCost:
    def __init__(self, prime_name, tuned, scaled):
        Formulae = MontgomeryIsogenyNew('svelu')
        self.curve = mont.MontgomeryCurve(prime_name)
        self.formulae = Formulae(self.curve, tuned, scaled)
        self.tuned = tuned
        self.Fp = self.curve.field
        self.ls = parameters['csidh'][prime_name]['L']
        self.p = parameters['csidh'][prime_name]['p']
        self.ls = parameters['csidh'][prime_name]['L']

    def get_torson_point(self, A, l):
        while True:
            Tp, _ = self.curve.elligator(A)
            Tp = self.curve.prac(self.curve.cofactor, Tp, A)
            for i in range(len(self.ls)):
                if not self.ls[i] == l:
                    Tp = self.curve.xmul(Tp, A, i)
            if not self.curve.isinfinity(Tp):
                return Tp

    def isog_sqrt(self, A, P, l):
        idx = self.formulae.L.index(l)

        # index system
        if self.tuned:
            self.formulae.set_parameters_velu(self.formulae.sJ_list[idx], self.formulae.sI_list[idx], idx)
        else:
            if l == 3:
                b = 0
                c = 0
            else:
                b = int(floor(sqrt(l - 1) / 2.0))
                c = int(floor((l - 1.0) / (4.0 * b)))
            self.formulae.set_parameters_velu(b, c, idx)

        T = self.get_torson_point(A, l)

        self.Fp.init_runtime()
        self.formulae.kps_s(T, A, idx)
        P = self.formulae.xeval_s(P, A)
        A = self.formulae.xisog_s(A, idx)

        return A, P, [self.Fp.fpmul, self.Fp.fpsqr, self.Fp.fpadd]

    def isog_trad(self, A, P, l):
        idx = self.formulae.L.index(l)
        T = self.get_torson_point(A, l)

        self.Fp.init_runtime()
        self.formulae.kps_t(T, A, idx)
        P = self.formulae.xeval_t(P, idx)
        A = self.formulae.xisog_t(A, idx)

        return A, P, [self.Fp.fpmul, self.Fp.fpsqr, self.Fp.fpadd]

    def isog_new(self, A, P, l):
        idx = self.formulae.L.index(l)

        # index system
        if self.tuned:
            self.formulae.set_parameters_velu(self.formulae.sJ_list[idx], self.formulae.sI_list[idx], idx)
        else:
            if l == 3:
                b = 0
                c = 0
            else:
                b = int(floor(sqrt(l - 1) / 2.0))
                c = int(floor((l - 1.0) / (4.0 * b)))
            self.formulae.set_parameters_velu(b, c, idx)

        T = self.get_torson_point(A, l)

        self.Fp.init_runtime()
        self.formulae.kps_s(T, A, idx)
        P = self.formulae.xeval_s(P, A)
        A = self.formulae.xisog_new(A, idx)

        return A, P, [self.Fp.fpmul, self.Fp.fpsqr, self.Fp.fpadd]

    def total_cost(cost):
        #return cost[0]*1.0+ cost[1]*0.8 + cost[2]*0.02
        return cost[0]*1.0+ cost[1]*1.0 + cost[2]*0.0

if __name__=="__main__":
    # system setting
    prime_name = 'p4096'
    tuned = True
    scaled = True

    isog = IsogenyCost(prime_name, tuned, scaled)
    isog_unscaled = IsogenyCost(prime_name, tuned, False)

    # a random point for evaluation
    x = isog.Fp(random.randint(2, isog.p))
    z = isog.Fp(random.randint(2, isog.p))
    P = [x, z]

    # starting curve
    a = isog.Fp(0)
    A = [a + 2, isog.Fp(4)]
    A, _, _= isog.isog_sqrt(A, P, 101)

    data = []
    for l in isog.ls:
        As, Ps, cost_s = isog.isog_sqrt(A, P, l)
        At, Pt, cost_t = isog.isog_trad(A, P, l)
        An, Pn, cost_n = isog.isog_new(A, P, l)
        As_uns, Ps_uns, cost_s_uns = isog_unscaled.isog_sqrt([isog_unscaled.Fp(x) for x in A], [isog_unscaled.Fp(x) for x in P], l)
        """
        print("l = %d, (M, S, a) = %s, %s, %s, total cost = %f, %f, %f, check: %s,%s"
            % (l, cost_s, cost_t, cost_n, IsogenyCost.total_cost(cost_s), IsogenyCost.total_cost(cost_t), IsogenyCost.total_cost(cost_n),
                    (As[0]/As[1] == At[0]/At[1] and Ps[0]/Ps[1] == Pt[0]/Pt[1]),
                    (As[0]/As[1] == An[0]/An[1] and Ps[0]/Ps[1] == Pn[0]/Pn[1])
                )
            )
        """
        ratio = IsogenyCost.total_cost(cost_n) / IsogenyCost.total_cost(cost_s_uns)
        data.append(ratio)
        if (As[0]/As[1] == At[0]/At[1] and Ps[0]/Ps[1] == Pt[0]/Pt[1]) and (As[0]/As[1] == An[0]/An[1] and Ps[0]/Ps[1] == Pn[0]/Pn[1]):
            print("l = %d, (M, S, a) = %s, %s, %s, total cost = %d, %d, %d, ration = %f"
                % (l, cost_s_uns, cost_t, cost_n,
                    IsogenyCost.total_cost(cost_s_uns), IsogenyCost.total_cost(cost_t), IsogenyCost.total_cost(cost_n),
                    ratio)
            )
    plt.scatter(isog.ls, data, color = "blue", marker = "+")
    plt.plot(isog.ls, [1 for l in isog.ls], color = "red")
    plt.xlabel('degree')
    plt.ylabel('(the cost of new formula) / (the cost of previous formula)');

    plt.savefig("costs.pdf")