from sibc.montgomery import MontgomeryIsogeny
from sibc.math import isequal, bitlength, hamming_weight, cswap

def MontgomeryIsogenyNew(name):
    formulae = MontgomeryIsogeny(name)

    class Formulae(formulae.func):
        def xisog_new(self, A, i):

            #A = [4*A[1], 4*A[1] - A[0]]

            AA = (A[0] + A[0])  # 2A' + 4C
            AA = (AA - A[1])  # 2A'
            AA = (AA + AA)  # 4A'

            # --------------------------------------------------------------------------------------------------
            #                   ~~~~~~~~
            #                    |    |
            # Computing E_J(W) = |    | [ F0(W, x([j]P)) * alpha^2 + F1(W, x([j]P)) * alpha + F2(W, x([j]P)) ]
            #                    j in J
            # In order to avoid costly inverse computations in fp, we are gonna work with projective coordinates
            # In particular, for a degree-l isogeny construction, we need alpha = 1 and alpha = -1

            # EJ_0 is the one determined by alpha = 0
            EJ_0 = [[0, 0, 0] for j in range(0, self.sJ, 1)]
            # EJ_1 is the one determined by alpha = 1
            EJ_1 = [[0, 0, 0] for j in range(0, self.sJ, 1)]

            for j in range(0, self.sJ, 1):

                # However, each self.SUB_SQUARED[j] and self.ADD_SQUARED[j] should be multiplied by C
                tadd = (self.ADD_SQUARED[j] * A[1])
                tsub = (self.SUB_SQUARED[j] * A[1])

                # We require the double of tadd and tsub
                tadd2 = (tadd + tadd)
                #tsub2 = (tsub + tsub)

                t1 = (self.XZJ4[j] * AA)  #       A *(-4*Xj*Zj)


                # Case alpha = 0

                EJ_0[j] = [0 - self.J[j][1], self.J[j][0]]

                # Case alpha = 1
                linear = (
                    t1 - tadd2
                )  #       A *(-4*Xj*Zj)  - C * (2 * (Xj + Zj)^2)
                EJ_1[j] = [tsub, linear, tsub]


            # The faster way for multiplying is using a divide-and-conquer approach
            poly_EJ_0 = self.poly_mul.product_tree(EJ_0, self.sJ)[
                'poly'
            ]  # product tree of EJ_0 (we only require the root)

            poly_EJ_1 = self.poly_mul.product_selfreciprocal_tree(EJ_1, self.sJ)[
                'poly'
            ]  # product tree of EJ_1 (we only require the root)

            if not self.SCALED_MULTIEVALUATION:

                """
                CAUTION
                If self.sI is not a power of two,
                then this unscaled remainder algorithm does not output the correct value.
                """

                remainders_EJ_0 = self.poly_redc.multieval_unscaled(
                    poly_EJ_0, self.sJ + 1, self.ptree_hI, self.sI
                )
                remainders_EJ_1 = self.poly_redc.multieval_unscaled(
                    poly_EJ_1, 2 * self.sJ + 1, self.ptree_hI, self.sI
                )


                const = 1
                const2 = 1

                if self.sI >= 4:

                    if self.sI > self.sJ:

                        const = self.ptree_hI['a'] ** ((bitlength(self.sI) - 3) * self.sI) * self.ptree_hI['poly'][self.sI]

                    else:

                        const = self.ptree_hI['a'] ** ((bitlength(self.sI) - 1) * self.sI) * self.ptree_hI['poly'][self.sI]

                elif self.L[i] > 3:

                    if self.sI > self.sJ:

                        const2 = self.ptree_hI['poly'][self.sI]

                    else:

                        const = self.ptree_hI['poly'][self.sI] ** 5
                        const2 = self.ptree_hI['a'] ** self.sI

            else:

                # Approach using scaled remainder trees
                if self.ptree_hI != None:
                    poly_EJ_0 = self.poly_redc.poly_redc(
                        poly_EJ_0, self.sJ + 1, self.ptree_hI
                    )
                    fg_0 = self.poly_mul.poly_mul_middle(
                        self.ptree_hI['scaled'], self.sI, poly_EJ_0[::-1], self.sI
                    )
                    remainders_EJ_0 = self.poly_redc.multieval_scaled(
                        fg_0[::-1],
                        self.sI,
                        [1] + [0] * (self.sI - 1),
                        self.sI,
                        self.ptree_hI,
                        self.sI,
                    )

                    poly_EJ_1 = self.poly_redc.poly_redc(
                        poly_EJ_1, 2 * self.sJ + 1, self.ptree_hI
                    )
                    fg_1 = self.poly_mul.poly_mul_middle(
                        self.ptree_hI['scaled'], self.sI, poly_EJ_1[::-1], self.sI
                    )
                    remainders_EJ_1 = self.poly_redc.multieval_scaled(
                        fg_1[::-1],
                        self.sI,
                        [1] + [0] * (self.sI - 1),
                        self.sI,
                        self.ptree_hI,
                        self.sI,
                    )
                else:
                    remainders_EJ_0 = []
                    remainders_EJ_1 = []


                const = 1
                const2 = 1

                if self.sI >= 3 and self.sJ > 1:

                    if self.sI > self.sJ:

                        const2 = self.ptree_hI['poly'][self.sI]

                    else:

                        const = self.ptree_hI['a'] ** (2 * self.sI)
                        const2 = self.ptree_hI['poly'][self.sI]


                elif self.sI == 3:

                    const = self.ptree_hI['a'] ** 3
                    const2 = self.ptree_hI['poly'][self.sI]


                elif self.sI == 2:

                    if self.sI > self.sJ:

                        const = self.ptree_hI['a'] ** 2
                        const2 = self.ptree_hI['poly'][self.sI] ** 3

                    else:

                        const = self.ptree_hI['poly'][self.sI] ** 3


                elif self.sI == 1:

                    const = self.ptree_hI['a']
                    const2 = self.ptree_hI['poly'][self.sI] * 2


            # Multipying all the remainders
            r0 = self.poly_mul.product(remainders_EJ_0, self.sI)
            r1 = self.poly_mul.product(remainders_EJ_1, self.sI)

            r0 = r0 ** 2 * A[1] ** (self.sI * self.sJ) * const2

            r1 = r1 * const


            # ---------------------------------------------------------------------------------
            # Now, we proceed by computing the missing part which is determined by K
            # Notice, the denominators are the same and then they annulled between them
            # In other words, it is not required to compute the product of all Zk's with k In K

            # Case alpha = 0
            hK_0 = [
                [self.K[k][0]]
                for k in range(0, self.sK, 1)
            ]
            hK_0 = self.poly_mul.product(
                hK_0, self.sK
            )  # product of (Xk) for each k in K
            # Case alpha = 1
            hK_1 = [
                [(self.K[k][1] - self.K[k][0])]
                for k in range(0, self.sK, 1)
            ]
            hK_1 = self.poly_mul.product(
                hK_1, self.sK
            )  # product of (Zk - Xk) for each k in K


            # --------------------------------------------------------------
            # Now, we have all the ingredients for computing the image curve

            t24 = 2 ** (4*self.L[i] - 4) * A[0] ** self.L[i] * (r0 * hK_0) ** 8
            t24m = A[1]**self.L[i] * (r1 * hK_1) ** 8

            return [t24, t24m]
    return Formulae
